% Implementation of Event Detection From Time Series Data
%
% Parameter:
%   e: likelihood threshold
%   MSet: set of models
%   data: the data series
%   window_size: parameter to optimize the runtime by using a sliding
%       window with the specified window size, pass zero if
%       you do not want to use sliding windows
%
% Return values:
%   score: the scores of the found change-points
%   changes: the change points
function [changes, scores] = eventDetection(e, MSet, data, window_size)
T = [];
scores = zeros(length(data),1);
changes = [];
global likelihood_cache;
likelihood_cache = NaN(length(data),1);

for i = 1:length(data)
    
    T(end+1,1) = data(i);
    if length(T) > 2 * (max(MSet)+2)
        
        %Sliding Window to increase speed if for a long time no change has
        %been detected
        if window_size > 0 && length(T) > window_size
            T = T(end-window_size+1:end);
        end
        split_likelihood_criteria = findSplitLikelihoodCriteria(T, MSet);
        no_split_likelihood_criteria = findLikelihoodCriteria(T, 0, MSet);
        cur_score = (no_split_likelihood_criteria - split_likelihood_criteria) / no_split_likelihood_criteria;
        scores(i) = cur_score;
        if cur_score > e
            T = [];
            changes(end+1,1) = i;
        end
    end
end
end

function [optimal_likelihood_criteria] = findSplitLikelihoodCriteria(T, MSet)
%Choose p to be greater than the highest polynomial degree + 2. This is
%necessary for the crossvalidation of polyfit. To calculate a unique
%polynomial the number of data points must be greater than the highest
%degree. Because in Leave-One-Out-Crossvalidation one value is left out, p
%must be as specified (I guess :-) )CS
p = max(MSet) + 2;
optimal_likelihood_criteria = Inf(1);

for i = p : size(T) - p - 1
    
    likelihood_criteria = findLikelihoodCriteria(T(1:i), p + i, MSet) + findLikelihoodCriteria(T(i+1:size(T)),0, MSet);
    if likelihood_criteria < optimal_likelihood_criteria;
        optimal_likelihood_criteria = likelihood_criteria;
    end
end
end

function [likelihood_criteria] = findLikelihoodCriteria(T, j, MSet)
%minimum_risk = Inf(1);
global likelihood_cache;
if (j ~= 0 && ~isnan(likelihood_cache(j)))
    likelihood_criteria = likelihood_cache(j);
else
    for i = 1:length(MSet)
        M = MSet(i);
        %model_risk = Risk(T, M);
        
        likelihood_criteria = Fit(T, M);
        if (j ~= 0)
            likelihood_cache(j) = likelihood_criteria;
        end
        
        % Not necessary
        %         if model_risk < minimum_risk
        %             minimum_risk = model_risk;
        %             likelihood_criteria = Fit(T, M);
        %         end
    end
end
end

function [risk] = Risk(T, M)
indices = 1:length(T);
indices = indices';
sse = 0;
for i = 1:5
    [train, test] = crossvalind('LeaveMOut',length(T),1);
    
    yhat = polyval(polyfit2(indices(train), T(train), M), indices(test));
    sse = sse + sum((yhat - T(test)).^2);
end
risk = sse / 5;
end

% Fit function to calculate sum of squared errors of prediction (sse).
% This sums the squared differences between the model and the data points.
function [fit] = Fit(T, M)
indices = 1:length(T);
indices = indices';
model = polyfit2(indices, T, M);
yhat = polyval(model,indices);

fit = sum((yhat - T).^2);
end
